@@ -4,9 +4,7 @@
  * @author tbellmann (Windows)
  * @author bernhard-thiele (Linux)
  * @since 2012-06-05
- * @copyright see Modelica_DeviceDrivers project's License.txt file
- *
- *
+ * @copyright see accompanying file LICENSE_Modelica_DeviceDrivers.txt
 */
 
 #include "../include/CompatibilityDefs.h"
@@ -20,7 +18,7 @@
 #include <malloc.h>
 #include "../../Include/MDDSpaceMouse.h"
 
-#define MAX_DEVICES 10
+#define MAX_DEVICES 50
 int SPC_nDevices = 0;
 /* Struct for the HID devices */
 typedef struct _HidDevice {
@@ -309,7 +307,7 @@ next:
  * @param[out] pdAxes array with 6 elements (value range [-1 1])
  * @param[out] piButtons array with 16 elements
  * */
-void MDD_spaceMouseGetData(double * pdAxes, int * piButtons) {
+DllExport void MDD_spaceMouseGetData(double * pdAxes, int * piButtons) {
     int i,j;
     if (SPC_ThreadID == 0) {
         CreateThread(0,1024,MDD_spaceMouseGetDataThread,NULL,0,&SPC_ThreadID);
